import eml_parser
from glob import glob
from sklearn.model_selection import train_test_split
from tqdm import tqdm
from data.AuthorshipDataset import AuthorshipDataset, LabelledCorpus
import numpy as np
from joblib import Parallel, delayed
from collections import Counter


class EnronMail(AuthorshipDataset):

    NUM_AUTHORS = 150
    MAX_MAIL_WORDS = 500
    TEST_SIZE = 0.1
    MIN_TOKENS = 10

    def __init__(self, data_path='../data/enron_mail_20150507/maildir/*', n_authors=-1, docs_by_author=-1, n_open_set_authors=0, random_state=42):
        self.data_path = data_path
        super().__init__(data_path, n_authors, docs_by_author, n_open_set_authors, random_state)

    def filter(self, base_str, filter_str):
        if filter_str in base_str:
            idx = base_str.index(filter_str)
            base_str = base_str[:idx]
        return base_str

    def _fetch_and_split(self):
        labels = []
        data = []

        path_list = self._get_most_prolific_authors(self.n_authors)
        emails_authors = Parallel(n_jobs=-1)(
            delayed(_fetch_emails_from_author)(author_path, EnronMail.MIN_TOKENS) for author_path in path_list
        )
        for emails, author in emails_authors:
            data.extend(emails)
            labels.extend([author]*len(emails))

        target_names = sorted(np.unique(labels))

        train_data, test_data, train_labels, test_labels = \
            train_test_split(data, labels, test_size=EnronMail.TEST_SIZE, stratify=labels)

        return LabelledCorpus(train_data, train_labels), LabelledCorpus(test_data, test_labels), target_names

    def _check_n_authors(self, n_authors, n_open_set_authors):
        pass

    def _get_most_prolific_authors(self, n):
        assert n <= EnronMail.NUM_AUTHORS, f'too many authors requested (maximum is {EnronMail.NUM_AUTHORS})'
        author_paths = glob(self.data_path)
        if n == -1:
            return author_paths
        author_count = Counter(
            {author_path :
                 len(glob(f'{author_path}/sent/*.')) +
                 len(glob(f'{author_path}/sent_items/*.'))
             for author_path in author_paths
             })
        return [path for path, count in author_count.most_common(n)]


def _fetch_emails_from_author(author_path, min_tokens):
    subject_filters = ['fw:', 'fwd:', 're:']
    body_filters = ['-----Original Message-----', '----- Forward', 'cc:', 'To:', 'to:', 'From:', 'from:']

    parsed_mails = 0
    author_mails = []
    author_docs = 0
    author_name = author_path[author_path.rindex('/') + 1:]
    author_bar = tqdm(list(glob(f'{author_path}/sent/*.')) + list(glob(f'{author_path}/sent_items/*.')))
    errors, trimmed = 0, 0
    for email in author_bar:
        author_bar.set_description(f'parsing for {author_path} errors={errors} trimmed={trimmed}')
        raw_email = open(email, 'rb').read()
        try:
            parsed_mail = eml_parser.eml_parser.decode_email_b(raw_email, include_raw_body=True)
            # subject = parsed_mail['header']['subject']
            body = parsed_mail['body'][0]['content']

            # for filter in subject_filters:
            #    if filter in subject.lower():
            #        continue

            # for filter in body_filters:
            #    body = self.filter(body, filter)

            # body = subject+'\n'+body
            body_tokens = body.split()
            ntokens = len(body_tokens)
            if ntokens >= min_tokens:
                if ntokens > EnronMail.MAX_MAIL_WORDS:
                    trimmed += 1
                    body = ' '.join(body_tokens[:EnronMail.MAX_MAIL_WORDS])
                author_mails.append(body)
                author_docs += 1
                # if n_docs_by_author != -1 and author_docs >= n_docs_by_author:
                #    add_author = True
                #    break

            parsed_mails += 1
        except Exception:
            errors += 1
            print(f'ERROR in file {email}')

    return author_mails, author_name
