import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.metrics import accuracy_score, f1_score
from tqdm import tqdm
import math
from sklearn.model_selection import train_test_split

from losses import SupConLoss1View, SupConLoss1ViewCrossEntropy
from model.early_stop import EarlyStop
from model.layers import FFProjection
from torch.utils.data import DataLoader


class AuthorshipAttributionClassifier(nn.Module):
    def __init__(self, projector, num_authors, pad_index, pad_length=500, device='cpu'):
        super(AuthorshipAttributionClassifier, self).__init__()
        self.projector = projector.to(device)
        self.ff = FFProjection(input_size=projector.output_size,
                               hidden_sizes=[],
                               output_size=num_authors).to(device)
        self.linear_proj = nn.Linear(projector.output_size, 128).to(device)  # to train the kernel alignment
        self.pad_index = pad_index
        self.pad_length = pad_length
        self.device = device

    def xavier_uniform(self):
        for p in self.parameters():
            if p.dim() > 1 and p.requires_grad:
                nn.init.xavier_uniform_(p)

    def fit(self, X, y, Xval, yval, batch_size, epochs, patience=20, lr=0.001, alpha=1., log='../log/tmp.csv', checkpointpath='../checkpoint/model.dat'):
        assert 0 <= alpha <= 1, 'wrong range, alpha must be in [0,1]'
        early_stop = EarlyStop(patience)

        #batcher = TwoClassBatch(batch_size=batch_size, n_epochs=epochs, steps_per_epoch=X.shape[0]//batch_size)
        criterion = torch.nn.CrossEntropyLoss().to(self.device)
        savcriterion = torch.nn.BCEWithLogitsLoss().to(self.device)
        optim = torch.optim.Adam(self.parameters(), lr=lr)

        tr_data = IndexedDataset(X, y, self.pad_length, self.pad_index, self.device)
        val_data = IndexedDataset(Xval, yval, self.pad_length, self.pad_index, self.device)

        with open(log, 'wt') as foo:
            print()
            foo.write('epoch\ttr-loss\tval-loss\tval-acc\tval-Mf1\tval-mf1\n')
            tr_loss = val_loss = acc = macrof1 = microf1 = -1
            pbar = tqdm(range(1, epochs + 1))
            for epoch in pbar:
                # training
                self.train()
                losses, attr_losses, sav_losses = [], [], []
                for xi, yi in tr_data.asDataLoader(batch_size, shuffle=True):
                    optim.zero_grad()
                    phi = self.projector(xi)

                    loss_attr = loss_sav = 0
                    loss_attr_value = loss_sav_value = -1

                    if alpha > 0:
                        logits = self.ff(phi)
                        loss_attr = criterion(logits, torch.as_tensor(yi).to(self.device))
                        loss_attr_value = loss_attr.item()

                    if alpha < 1:

                        # choose balanced number of positive (same author) and negative (different authors)
                        idx1, idx2, sav_labels = choose_sav_pairs(yi, npairs=batch_size)

                        phi1 = phi[idx1]
                        phi2 = phi[idx2]
                        cross = torch.bmm(phi1.unsqueeze(1), phi2.unsqueeze(2).permute(0,1,2)).squeeze()
                        loss_sav = savcriterion(cross.unsqueeze(0), torch.as_tensor(sav_labels).float().unsqueeze(0).to(self.device))
                        loss_sav_value = loss_sav.item()

                        # add a cross-entropy based criterion (instead of KTA -- let's see how it works)

                        ## todo: optimize (only upper diagonal)
                        #kernel = torch.matmul(phi, phi.T)
                        #ideal_kernel = torch.as_tensor(1 * (np.outer(1 + yi, 1 / (yi + 1)) == 1)).to(self.device)
                        # todo: maybe the KALoss should take into consideration the balance (it is more likely to have
                        # a pair of negative examples than positives)
                        #loss_sav = KernelAlignmentLoss(kernel, ideal_kernel)
                        #loss_sav_value = loss_sav.item()

                    loss = loss_attr*alpha + loss_sav*(1.-alpha)

                    loss.backward()
                    optim.step()

                    attr_losses.append(loss_attr_value)
                    sav_losses.append(loss_sav_value)
                    losses.append(loss.item())
                    tr_loss = np.mean(losses)
                    pbar.set_description(f'training epoch={epoch} '
                                         f'loss={tr_loss:.5f} '
                                         f'attr-loss={np.mean(attr_losses):.5f} '
                                         f'sav-loss={np.mean(sav_losses):.5f} '
                                         f'val_loss={val_loss:.5f} val_acc={acc:.4f} macrof1={macrof1:.4f} microf1={microf1:.4f} '
                                         f'patience={early_stop.patience}/{early_stop.patience_limit}')

                # validation
                self.eval()
                with torch.no_grad():
                    predictions, losses = [], []
                    for xi, yi in val_data.asDataLoader(batch_size, shuffle=False):
                        logits = self.forward(xi)
                        loss = criterion(logits, torch.as_tensor(yi).to(self.device))
                        losses.append(loss.item())
                        logits = nn.functional.log_softmax(logits, dim=1)
                        prediction = tensor2numpy(torch.argmax(logits, dim=1).view(-1))
                        predictions.append(prediction)
                    val_loss = np.mean(losses)
                    predictions = np.concatenate(predictions)
                    acc = accuracy_score(yval, predictions)
                    macrof1 = f1_score(yval, predictions, average='macro')
                    microf1 = f1_score(yval, predictions, average='micro')

                    foo.write(f'{epoch}\t{tr_loss:.8f}\t{val_loss:.8f}\t{acc:.3f}\t{macrof1:.3f}\t{microf1:.3f}\n')
                    foo.flush()

                early_stop(microf1, epoch)
                if early_stop.IMPROVED:
                    torch.save(self.state_dict(), checkpointpath)
                elif early_stop.STOP:
                    break
        print(f'training ended; loading best model parameters in {checkpointpath} for epoch {early_stop.best_epoch}')
        self.load_state_dict(torch.load(checkpointpath))
        return early_stop.best_score

    def supervised_contrastive_learning(self, X, y, Xval, yval, batch_size, epochs, patience=200, lr=0.001, log='../log/tmp.csv', checkpointpath='../checkpoint/model.dat'):
        early_stop = EarlyStop(patience, lower_is_better=True)

        criterion = SupConLoss1View().to(self.device)
        optim = torch.optim.Adam(self.parameters(), lr=lr)  # take only phi's parameters

        tr_data = IndexedDataset(X, y, self.pad_length, self.pad_index, self.device)
        val_data = IndexedDataset(Xval, yval, self.pad_length, self.pad_index, self.device)

        with open(log, 'wt') as foo:
            print()
            foo.write('epoch\ttr-loss\tval-loss\tval-acc\tval-Mf1\tval-mf1\n')
            tr_loss, val_loss, neg_losses_val, pos_losses_val = -1, -1, -1, -1
            pbar = tqdm(range(1, epochs + 1))
            for epoch in pbar:
                # training
                self.train()
                losses, pos_losses, neg_losses = [], [], []
                for xi, yi in tr_data.asDataLoader(batch_size, shuffle=True):
                    optim.zero_grad()
                    phi = self.projector(xi)
                    phi = self.linear_proj(phi)
                    phi = F.normalize(phi, p=2, dim=-1)
                    contrastive_loss, neg_loss, pos_loss = criterion(phi, torch.as_tensor(yi).to(self.device))
                    contrastive_loss.backward()
                    optim.step()
                    losses.append(contrastive_loss.item())
                    neg_losses.append(neg_loss.item())
                    pos_losses.append(pos_loss.item())
                    tr_loss = np.mean(losses)

                    pbar.set_description(f'training epoch={epoch} '
                                         f'loss={tr_loss:.5f} [neg={np.mean(neg_losses):.5f}, pos={np.mean(pos_losses):.5f}] '
                                         f'val_loss={val_loss:.5f} [neg={np.mean(neg_losses_val):.5f}, pos={np.mean(pos_losses_val):.5f}] '
                                         f'patience={early_stop.patience}/{early_stop.patience_limit}')

                # validation
                self.eval()
                with torch.no_grad():
                    losses, pos_losses_val, neg_losses_val = [], [], []
                    for xi, yi in val_data.asDataLoader(batch_size, shuffle=False):
                        phi = self.projector(xi)
                        contrastive_loss, neg_loss, pos_loss = criterion(phi, torch.as_tensor(yi).to(self.device))
                        losses.append((neg_loss + pos_loss).item())
                        neg_losses_val.append(neg_loss.item())
                        pos_losses_val.append(pos_loss.item())
                    val_loss = np.mean(losses)

                early_stop(val_loss, epoch)
                if early_stop.IMPROVED:
                    torch.save(self.state_dict(), checkpointpath)
                elif early_stop.STOP:
                    break
        print(f'training ended; loading best model parameters in {checkpointpath} for epoch {early_stop.best_epoch}')
        self.load_state_dict(torch.load(checkpointpath))
        return early_stop.best_score

    def train_linear_classifier(self, X, y, Xval, yval, batch_size, epochs, patience=25, lr=0.001, log='../log/tmp.csv', checkpointpath='../checkpoint/model.dat'):
        early_stop = EarlyStop(patience)

        criterion = torch.nn.CrossEntropyLoss().to(self.device)
        optim = torch.optim.Adam(self.parameters(), lr=lr)

        tr_data = IndexedDataset(X, y, self.pad_length, self.pad_index, self.device)
        val_data = IndexedDataset(Xval, yval, self.pad_length, self.pad_index, self.device)

        tr_loss = val_loss = acc = macrof1 = microf1 = -1
        pbar = tqdm(range(1, epochs + 1))
        for epoch in pbar:
            # training
            self.train()
            losses = []
            for xi, yi in tr_data.asDataLoader(batch_size, shuffle=True):

                with torch.no_grad():
                    phi = self.projector(xi)
                logits = self.ff(phi.detach())

                optim.zero_grad()
                loss = criterion(logits, torch.as_tensor(yi).to(self.device))
                loss.backward()
                optim.step()

                losses.append(loss.item())
                tr_loss = np.mean(losses)
                pbar.set_description(f'training epoch={epoch} '
                                     f'loss={tr_loss:.5f} '
                                     f'val_loss={val_loss:.5f} val_acc={acc:.4f} macrof1={macrof1:.4f} microf1={microf1:.4f} '
                                     f'patience={early_stop.patience}/{early_stop.patience_limit}')

            # validation
            self.eval()
            with torch.no_grad():
                predictions, losses = [], []
                for xi, yi in val_data.asDataLoader(batch_size, shuffle=False):
                    logits = self.forward(xi)
                    loss = criterion(logits, torch.as_tensor(yi).to(self.device))
                    losses.append(loss.item())
                    logits = nn.functional.log_softmax(logits, dim=1)
                    prediction = tensor2numpy(torch.argmax(logits, dim=1).view(-1))
                    predictions.append(prediction)
                val_loss = np.mean(losses)
                predictions = np.concatenate(predictions)
                acc = accuracy_score(yval, predictions)
                macrof1 = f1_score(yval, predictions, average='macro')
                microf1 = f1_score(yval, predictions, average='micro')

            early_stop(microf1, epoch)
            if early_stop.IMPROVED:
               torch.save(self.state_dict(), checkpointpath)
            elif early_stop.STOP:
               break
        print(f'training ended; loading best model parameters in {checkpointpath} for epoch {early_stop.best_epoch}')
        self.load_state_dict(torch.load(checkpointpath))
        return early_stop.best_score

    def predict(self, x, batch_size=100):
        self.eval()
        te_data = IndexedDataset(x, None, self.pad_length, self.pad_index, self.device)
        predictions = []
        with torch.no_grad():
            for xi in te_data.asDataLoader(batch_size, shuffle=False):
                logits = self.forward(xi)
                logits = nn.functional.log_softmax(logits, dim=1)
                prediction = tensor2numpy(torch.argmax(logits, dim=1).view(-1))
                predictions.append(prediction)
        return np.concatenate(predictions)

    def project(self, x, batch_size=100):
        self.eval()
        te_data = IndexedDataset(x, None, self.pad_length, self.pad_index, self.device)
        predictions = []
        with torch.no_grad():
            for xi in te_data.asDataLoader(batch_size, shuffle=False):
                phi = tensor2numpy(self.projector(xi))
                predictions.append(phi)
        return np.concatenate(predictions)

    def project_kernel(self, x, batch_size=100):
        self.eval()
        te_data = IndexedDataset(x, None, self.pad_length, self.pad_index, self.device)
        predictions = []
        with torch.no_grad():
            for xi in te_data.asDataLoader(batch_size, shuffle=False):
                phi = self.projector(xi)
                phi = self.linear_proj(phi)
                phi = F.normalize(phi, p=2, dim=-1)
                phi = tensor2numpy(phi)
                predictions.append(phi)
        return np.concatenate(predictions)


    def forward(self, x):
        phi = self.projector(x)
        return self.ff(phi)


def choose_sav_pairs(y, npairs):
    n = len(y)
    y = y+1  # reindex from [0..n_classes-1] to [1..n_classes] for convenience
    same_author = (np.outer(y, 1/y) == 1)
    triu = np.triu_indices(n, k=1)
    same_author_nodup = same_author[triu]
    idxi, idxj = triu

    posi, negi = idxi[same_author_nodup], idxi[same_author_nodup == False]
    posj, negj = idxj[same_author_nodup], idxj[same_author_nodup == False]
    num_pos = same_author_nodup.sum()
    num_neg = len(same_author_nodup)-num_pos  # == len(posj)

    # balanced:
    pos_take = np.random.choice(np.arange(num_pos), npairs//2, replace=num_pos < npairs//2)
    posi, posj = posi[pos_take], posj[pos_take]

    neg_take = np.random.choice(np.arange(num_neg), npairs//2, replace=num_neg < npairs//2)
    negi, negj = negi[neg_take], negj[neg_take]

    idx1 = np.concatenate([posi, negi])
    idx2 = np.concatenate([posj, negj])
    savlabels = np.array([1]*len(posi) + [0]*len(negi))

    return idx1, idx2, savlabels






# class SameAuthorClassifier(nn.Module):
#     def __init__(self, projector, num_authors, pad_index, pad_length=500, device='cpu'):
#         super(SameAuthorClassifier, self).__init__()
#         self.projector = projector.to(device)
#         self.padder = Padding(pad_index=pad_index, max_length=pad_length, dynamic=True, pad_at_end=False, device=device)
#         self.device = device
#
#     def fit(self, X, y, batch_size, epochs, lr=0.001, steps_per_epoch=100):
#         self.train()
#         batcher = TwoClassBatch(batch_size=batch_size, n_epochs=epochs, steps_per_epoch=steps_per_epoch)
#         optim = torch.optim.Adam(self.parameters(), lr=lr)
#
#         pbar = tqdm(range(batcher.n_epochs))
#         for epoch in pbar:
#             losses = []
#             for xi, yi in batcher.epoch(X, y):
#                 optim.zero_grad()
#                 xi = self.padder.transform(xi)
#                 phi = self.projector(xi)
#                 #normalize phi to have norm 1? maybe better as the last step of projector
#                 kernel = torch.matmul(phi, phi.T)
#                 ideal_kernel = torch.as_tensor(1 * (np.outer(1 + yi, 1 / (yi + 1)) == 1)).to(self.device)
#                 loss = KernelAlignmentLoss(kernel, ideal_kernel)
#                 loss.backward()
#                 #clip_gradient(model)
#                 optim.step()
#                 losses.append(loss.item())
#                 pbar.set_description(f'training epoch={epoch} loss={np.mean(losses):.5f}')
#
#     def predict(self, x, z, batch_size=100):
#         self.eval()
#         batcher = Batch(batch_size=batch_size, n_epochs=1, shuffle=False)
#         predictions = []
#         for xi, zi in tqdm(batcher.epoch(x, z), desc='test'):
#             xi = self.padder.transform(xi)
#             zi = self.padder.transform(zi)
#             inners = self.forward(xi, zi)
#             prediction = tensor2numpy(inners) > 0.5 # is this correct? should it be > 0 and the ideal kernel in field {-1,+1}?
#             predictions.append(prediction)
#         return np.concatenate(predictions)
#
#     def forward(self, x, z):
#         assert x.shape == z.shape, 'shape mismatch between matrices x and z'
#         phi_x = self.projector(x)
#         phi_z = self.projector(z)
#         rows, cols = phi_x.shape
#         pairwise_inners = torch.bmm(phi_x.view(rows, 1, cols), phi_z.view(rows, cols, 1)).squeeze()
#         return pairwise_inners


# class FullAuthorClassifier(nn.Module):
#     def __init__(self, projector, num_authors, pad_index, pad_length=500, device='cpu'):
#         super(FullAuthorClassifier, self).__init__()
#         self.projector = projector.to(device)
#         self.ff = FFProjection(input_size=projector.space_dimensions(),
#                                hidden_sizes=[1024],
#                                output_size=num_authors).to(device)
#         self.padder = Padding(pad_index=pad_index, max_length=pad_length, dynamic=True, pad_at_end=False, device=device)
#         self.device = device
#
#     def fit(self, X, y, batch_size, epochs, lr=0.001, steps_per_epoch=100):
#         self.train()
#         batcher = TwoClassBatch(batch_size=batch_size, n_epochs=epochs, steps_per_epoch=steps_per_epoch)
#         criterion = torch.nn.CrossEntropyLoss().to(self.device)
#         optim = torch.optim.Adam(self.parameters(), lr=lr)
#         alpha = 0.5
#
#         pbar = tqdm(range(batcher.n_epochs))
#         for epoch in pbar:
#             losses, sav_losses, attr_losses = [], [], []
#             for xi, yi in batcher.epoch(X, y):
#                 optim.zero_grad()
#                 xi = self.padder.transform(xi)
#                 phi = self.projector(xi)
#                 #normalize phi to have norm 1? maybe better as the last step of projector
#
#                 #sav-loss
#                 kernel = torch.matmul(phi, phi.T)
#                 ideal_kernel = torch.as_tensor(1 * (np.outer(1 + yi, 1 / (yi + 1)) == 1)).to(self.device)
#                 sav_loss = KernelAlignmentLoss(kernel, ideal_kernel)
#                 sav_losses.append(sav_loss.item())
#
#                 #attr-loss
#                 logits = self.ff(phi)
#                 attr_loss = criterion(logits, torch.as_tensor(yi).to(self.device))
#                 attr_losses.append(attr_loss.item())
#
#                 #loss
#                 loss = (alpha)*sav_loss + (1-alpha)*attr_loss
#                 losses.append(loss.item())
#
#                 loss.backward()
#                 #clip_gradient(model)
#                 optim.step()
#                 pbar.set_description(
#                     f'training epoch={epoch} '
#                     f'sav-loss={np.mean(sav_losses):.5f} '
#                     f'attr-loss={np.mean(attr_losses):.5f} '
#                     f'loss={np.mean(losses):.5f}'
#                 )
#
#     def predict_sav(self, x, z, batch_size=100):
#         self.eval()
#         batcher = Batch(batch_size=batch_size, n_epochs=1, shuffle=False)
#         predictions = []
#         for xi, zi in tqdm(batcher.epoch(x, z), desc='test'):
#             xi = self.padder.transform(xi)
#             zi = self.padder.transform(zi)
#             phi_xi = self.projector(xi)
#             phi_zi = self.projector(zi)
#             rows, cols = phi_xi.shape
#             pairwise_inners = torch.bmm(phi_xi.view(rows, 1, cols), phi_zi.view(rows, cols, 1)).squeeze()
#             prediction = tensor2numpy(pairwise_inners) > 0.5 # is this correct? should it be > 0 and the ideal kernel in field {-1,+1}?
#             predictions.append(prediction)
#         return np.concatenate(predictions)
#
#     def predict_labels(self, x, batch_size=100):
#         self.eval()
#         batcher = Batch(batch_size=batch_size, n_epochs=1, shuffle=False)
#         predictions = []
#         for xi in tqdm(batcher.epoch(x), desc='test'):
#             xi = self.padder.transform(xi)
#             phi = self.projector(xi)
#             logits = self.ff(phi)
#             prediction = tensor2numpy( torch.argmax(logits, dim=1).view(-1))
#             predictions.append(prediction)
#         return np.concatenate(predictions)

#def KernelAlignmentLoss(K, Y):
#    n_el = K.shape[0]*K.shape[1]
#    loss = torch.norm(K - Y, p='fro')  # in Nello's paper this is different
#    loss = loss / n_el  # this is in order to factor out the accumulation which is only due to the size
#    return loss


# class TwoClassBatch:
#     """
#     given a X and y (multi-label) produces batches of elements of X, y for two classes (e.g., c1, c2)
#     of equal size, i.e., the batch is [(x1,c1), ..., (xn,c1), (xn+1,c2), ..., (x2n,c2)]
#     """
#     def __init__(self, batch_size, n_epochs, steps_per_epoch):
#         self.batch_size = batch_size
#         self.n_epochs = n_epochs
#         self.steps_per_epoch = steps_per_epoch
#         self.current_epoch = 0
#         if self.batch_size % 2 != 0:
#             raise ValueError('warning, batch size is not even')
#
#     def epoch(self, X, y):
#         n_el = len(y)
#         assert X.shape[0] == n_el, 'inconsistent sizes in X, y'
#         classes = np.unique(y)
#         groups = {ci: X[y==ci] for ci in classes}
#         class_prevalences = [len(groups[ci])/n_el for ci in classes]
#         n_choices = self.batch_size // 2
#
#         for b in range(self.steps_per_epoch):
#             class1, class2 = np.random.choice(classes, p=class_prevalences, size=2, replace=False)
#             X1 = np.random.choice(groups[class1], size=n_choices)
#             X2 = np.random.choice(groups[class2], size=n_choices)
#             X_batch = np.concatenate([X1,X2])
#             y_batch = np.repeat([class1, class2], repeats=[n_choices,n_choices])
#             yield X_batch, y_batch
#         self.current_epoch += 1


def tensor2numpy(t):
    return t.to('cpu').detach().numpy()


# ------------

class IndexedDataset(torch.utils.data.Dataset):
    def __init__(self, X, y, MAX_LENGTH, padindex, device, pad_at_end=False):
        self.X = X
        self.y = y
        self.MAX_LENGTH = MAX_LENGTH
        self.padindex = padindex
        self.device = device
        self.pad_at_end = pad_at_end

    def __len__(self):
        return len(self.X)
    
    @property
    def islabelled(self):
        return self.y is not None

    def __getitem__(self, index):
        if self.islabelled:
            return self.X[index], self.y[index]
        else:
            return self.X[index]
        
    def collate_pad_fn(self, batch):
        """
        :param batch: a list of lists of indexes (integers)
        :return: a torch.tensor of shape (n,m) where n is the number of elements in X_batch and m is the pad length
        (the maximum in elements of X_batch)
        """
        if self.islabelled:
            X, y = list(zip(*batch))
        else:
            X = batch
        lengths = list(map(len, X))
        pad_length = min(max(lengths), self.MAX_LENGTH)
        X = [x[:pad_length] for x in X]
        if self.pad_at_end:
            padded = [x + [self.padindex] * (pad_length - x_len) for x, x_len in zip(X, lengths)]
        else:
            padded = [[self.padindex] * (pad_length - x_len) + x for x, x_len in zip(X, lengths)]

        X = torch.from_numpy(np.asarray(padded, dtype=int)).to(self.device)
        if self.islabelled:
            y = torch.from_numpy(np.asarray(y)).to(self.device)
            return X, y
        else:
            return X

    def asDataLoader(self, batch_size, shuffle):
        return torch.utils.data.DataLoader(self, batch_size=batch_size, shuffle=shuffle, collate_fn=self.collate_pad_fn)
