# adapted from https://github.com/Shawn1993/cnn-text-classification-pytorch/blob/master/model.py
import torch
import torch.nn as nn
import torch.nn.functional as F


class Phi(nn.Module):

    def __init__(self, cnn, ff):
        super(Phi, self).__init__()
        self.cnn = cnn
        self.ff = ff
        self.output_size = self.ff.output_size

    def forward(self, x):
        x = self.cnn(x)
        x = self.ff(x)
        x = F.normalize(x, p=2, dim=-1)
        return x


class CNNProjection(nn.Module):

    def __init__(self, vocabulary_size, embedding_dim, channels_out, kernel_sizes):
        super(CNNProjection, self).__init__()
        channels_in = 1
        self.embed = nn.Embedding(vocabulary_size, embedding_dim)
        self.convs1 = nn.ModuleList(
            [nn.Conv2d(channels_in, channels_out, (K, embedding_dim)) for K in kernel_sizes]
        )
        self.output_size = len(kernel_sizes) * channels_out

    def convolve(self, x):
        x = x.unsqueeze(1)  # (N, Ci, W, D)
        x = [self.conv_and_pool(x, conv) for conv in self.convs1]  # [(N, Co), ...]*len(Ks)
        x = torch.cat(x, 1)
        return x

    def conv_and_pool(self, x, conv):
        x = F.relu(conv(x)).squeeze(3)  # (N, Co, W)
        x = F.max_pool1d(x, x.size(2)).squeeze(2)
        return x

    def forward(self, x):
        x = self.embed(x)  # (N, W, D)
        x = self.convolve(x)  # (N, len(Ks)*Co]
        return x


class FFProjection(nn.Module):
    def __init__(self, input_size, hidden_sizes, output_size, activation=nn.functional.relu, dropout=0.5,
                 activate_last=False):
        super(FFProjection, self).__init__()
        sizes = [input_size] + hidden_sizes + [output_size]
        self.ff = nn.ModuleList([
            nn.Linear(sizes[i], sizes[i+1]) for i in range(len(sizes)-1)
        ])
        self.activation = activation
        self.dropout = nn.Dropout(p=dropout)
        self.activate_last = activate_last
        self.output_size = output_size

    def forward(self, x):
        last_layer_idx = len(self.ff)-1
        for i, linear in enumerate(self.ff):
            x = linear(x)
            if i < last_layer_idx or self.activate_last:
                x = self.dropout(self.activation(x))
        return x


# deprecated
# class RNNProjection(nn.Module):
#     def __init__(self, vocab_size, hidden_size, output_size, device='cpu'):
#         super(RNNProjection, self).__init__()
#         self.output_size = output_size
#         self.hidden_size = hidden_size
#         self.vocab_size = vocab_size
#         self.num_layers=1
#         self.num_directions=1
#         self.device = device
#
#         self.embedding = nn.Embedding(vocab_size, hidden_size).to(device)
#         self.rnn = nn.GRU(
#             input_size=hidden_size,
#             hidden_size=hidden_size,
#             num_layers=self.num_layers,
#             bidirectional=(self.num_directions == 2),
#             batch_first=True
#         ).to(device)
#         self.projection = nn.Linear(self.num_layers * self.num_directions * self.hidden_size, output_size).to(device)
#
#     def init_hidden(self, batch_size):
#         return torch.zeros(self.num_layers * self.num_directions, batch_size, self.hidden_size).to(self.device)
#
#     def forward(self, x):
#         batch_size = x.shape[0]
#         x = self.embedding(x)
#         output, hn = self.rnn(x, self.init_hidden(batch_size))
#         hn = hn.view(self.num_layers, self.num_directions, batch_size, self.hidden_size)
#         hn = hn.permute(2, 0, 1, 3).reshape(batch_size, -1)
#         return self.projection(hn)
#
#     def space_dimensions(self):
#         return self.output_size